/* ldap.c - LDAP access
 *      Copyright (C) 2002 Klarlvdalens Datakonsult AB
 *
 * This file is part of DirMngr.
 *
 * DirMngr is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DirMngr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>

#include <string.h>
#include <ldap.h>

#include "crlfetch.h"
#include "../assuan/assuan.h"
#include "dirmngr.h"


#define CERTATTR "userCertificate;binary"
#define CACERTATTR "caCertificate;binary"

/*! Fetch the value attr, put result in value, length in valuelen.
 *  Returns 0 on succes
 */
static int 
get_attr_from_result_ldap( LDAP* ld, LDAPMessage* msg, const char* attr, 
				char** value, size_t* valuelen )
/* FIXME: It is assumed that that attr is a null terminated char string. */
{
  LDAPMessage* item = ldap_first_entry( ld, msg );  
  struct berval** values;
  if( !item ) return -1;
  values = ldap_get_values_len( ld, item, attr );

  /* try without ";binary" .. which worked with some strange LDAP servers. 
   * Note: Strings are hardcoded to avoid fragile string operations. */
  if (!values) {
    if (strncasecmp(attr,"certificateRevocationList;binary",40)==0)
      values = ldap_get_values_len( ld, item, "certificateRevocationList");
    if (strncasecmp(attr,"cACertificate;binary",40)==0)
      values = ldap_get_values_len( ld, item, "cACertificate");
    if (strncasecmp(attr,"userCertificate;binary",40)==0)
      values = ldap_get_values_len( ld, item, "userCertificate");
    if (strncasecmp(attr,"authorityRevocationList;binary",40)==0)
      values = ldap_get_values_len( ld, item, "authorityRevocationList");
  }

  if( values ) {
    *valuelen = values[0]->bv_len;
    *value = xmalloc( *valuelen * sizeof( char ) );
    memcpy( *value, values[0]->bv_val, *valuelen );
    ldap_value_free_len( values );
    return ASSUAN_No_Error;
  } else return ASSUAN_No_CRL_For_Cert;
}

static int
attr_fetch_ldap_internal( const char* host, int port, 
			  const char* user, const char* pass, 
			  const char* dn, const char* attr, 
			  char** value, size_t* valuelen )
{
  char* attrs[] = {(char*)attr, NULL};
  int rc;
  int msgid;

  LDAPMessage* result;
  LDAP* ld;

  log_debug( "trying to fetch attr %s from \"%s\" / \"%s\"\n", attr, host, dn );

  ld = ldap_init( host, port );
  if( !ld ) {
    perror( "Error in ldap_init" );
    return ASSUAN_No_CRL_For_Cert;
  }
  msgid = ldap_simple_bind( ld, user, pass );
  /*if( ldap_simple_bind_s( ld, user, pass ) != LDAP_SUCCESS ) {*/
  rc = ldap_result( ld, msgid, 0, &dirmngr_opt.ldaptimeout, &result );
  if( rc != LDAP_RES_BIND ) {
    ldap_perror( ld, "Error in ldap_result" );
    return ASSUAN_No_CRL_For_Cert;
  }
  
  if( ldap_search_st( ld, dn, LDAP_SCOPE_BASE, "objectclass=*", attrs , 0, &dirmngr_opt.ldaptimeout, 
		      &result ) 
      != LDAP_SUCCESS ) {
    log_debug( "Error in line %s:%d\n", __FILE__, __LINE__ );
    ldap_perror( ld, "Error during ldap_seach_st()" );
    log_debug( "dn=\"%s\"\nattr=\"%s\"\n" , dn, attr);    
    return ASSUAN_No_CRL_For_Cert;
  }

  rc = get_attr_from_result_ldap( ld, result, attr, value, valuelen );
  if( rc ) {
    log_debug( "Error in line %s:%d\n", __FILE__, __LINE__ );
    ldap_perror( ld, "Error during get_attr_from_result()" );
    return ASSUAN_No_CRL_For_Cert;
  }
  ldap_msgfree(result);
  return ASSUAN_No_Error;
}

int 
attr_fetch_ldap(const char* dn,  const char* attr, char** value, size_t* valuelen )
{
  struct ldap_server_s* servers = dirmngr_opt.ldapservers;
  int rc = ASSUAN_No_CRL_For_Cert;
  while( servers != NULL ) {
    
    rc = attr_fetch_ldap_internal( servers->host, servers->port, 
				   servers->user, servers->pass, 
				   dn, attr,
				   value, valuelen );
    if( rc == ASSUAN_No_Error ) break;
    servers = servers->next;
  }
  return rc;
}

static int 
url_fetch_ldap_internal( const LDAPURLDesc* ludp, const char* attr, 
			 char** value, size_t* valuelen )
{
  LDAPMessage* result;
  LDAP* ld;
  int rc;

  log_debug( "url_fetch_ldap_internal(), attr = %s, ludp->lud_attrs = %s", attr, 
	     ludp->lud_attrs?ludp->lud_attrs[0]:"(null)"  );

  ld = ldap_init( ludp->lud_host, ludp->lud_port );

  if( !ld ) {
    perror( "Error in ldap_init" );
    return ASSUAN_No_CRL_For_Cert;
  }
  if( ldap_simple_bind_s( ld, 0, 0 ) != LDAP_SUCCESS ) {
    ldap_perror( ld, "Error in ldap_simple_bind_s" );
    return ASSUAN_No_CRL_For_Cert;
  }
  
  /*
    int ldap_search_s(ld, base, scope, filter, attrs, attrsonly, res)
  */
  if( ldap_search_st( ld, ludp->lud_dn, ludp->lud_scope, ludp->lud_filter, ludp->lud_attrs, 0, 
		      &dirmngr_opt.ldaptimeout, &result ) 
      != LDAP_SUCCESS ) {
    log_debug( "Error in line %s:%d\n", __FILE__, __LINE__ );
    ldap_perror( ld, "Error during ldap_search_s()" );
    return ASSUAN_No_CRL_For_Cert;
  }

  if( ludp->lud_attrs != NULL ) {
    rc = get_attr_from_result_ldap( ld, result, ludp->lud_attrs[0], value, valuelen );
  } else { 
    rc = get_attr_from_result_ldap( ld, result, attr, value, valuelen );
  }
  if( rc ) {
    log_debug( "Error in line %s:%d\n", __FILE__, __LINE__ );
    ldap_perror( ld, "Error during get_attr_from_result()" );
    return ASSUAN_No_CRL_For_Cert;
  }
  ldap_msgfree(result);  
  return ASSUAN_No_Error;
}

static void add_server_to_servers( const char* host, int port )
{
  struct ldap_server_s* servers = dirmngr_opt.ldapservers;
  struct ldap_server_s* end = dirmngr_opt.ldapservers;
  struct ldap_server_s* s;
  while( servers ) {
    if( !strcmp( servers->host, host ) && servers->port == port ) {
      /* already in list... */
      return;
    }
    end = servers;
    servers = servers->next;
  }
  log_debug( "added %s:%d to server-list\n", host, port );
  s = xmalloc( sizeof( struct ldap_server_s ) );
  s->host = xstrdup( host );
  s->port = port;
  s->user = 0;
  s->pass = 0;
  s->base = 0;
  s->next = 0;
  end->next = s;
}

int 
url_fetch_ldap(const char* url, const char* attr, char** value, size_t* valuelen )
{
  LDAPURLDesc* ludp = 0;
  int rc = ASSUAN_No_CRL_For_Cert;

  log_debug( "Looking up url %s\n", url );
  if( !ldap_is_ldap_url( url ) ) {
    log_debug( "%s is not an ldap url\n", url );
    *value = 0;
    *valuelen = 0;
    return ASSUAN_No_CRL_For_Cert; /* PENDING(steffen) better error */
  }

  if( ldap_url_parse( url, &ludp ) ) {
    log_debug( "invalid ldap url %s\n", url );
    *value = 0;
    *valuelen = 0;
    return ASSUAN_No_CRL_For_Cert; // PENDING(steffen) better error    
  }
  log_debug( "trying to fetch attr %s from %s\n", attr, url );
  if( !ludp->lud_host ) {
    struct ldap_server_s* servers = dirmngr_opt.ldapservers;
    char* oldhost = ludp->lud_host;
    log_debug("No hostname in URL, using defaults\n" );
    while( servers ) {
      ludp->lud_host = servers->host;
      ludp->lud_port = servers->port;      
      rc = url_fetch_ldap_internal( ludp, attr, value, valuelen );
      if( rc == ASSUAN_No_Error ) break;
      servers = servers->next;
    }
    ludp->lud_host = oldhost;
    ldap_free_urldesc( ludp );
    return rc;
  } else {
    /* URL is good, just query */
    if( dirmngr_opt.add_new_ldapservers ) add_server_to_servers( ludp->lud_host, ludp->lud_port );
    rc = url_fetch_ldap_internal( ludp, attr, value, valuelen );
    ldap_free_urldesc( ludp );
    return rc;
  }

}

/*
 * Possible patterns:
 *
 *   KeyID
 *   Fingerprint
 *   OpenPGP userid
 * x <email@somewhere.org>
 *   Exact word match in user id or subj. name
 * x Subj. DN
 *   Issuer DN
 *   Serial number + subj. DN
 * x Substring match
 */
 
char*
parse_one_pattern( const char* pattern, int* base )
{
  const char* p = pattern;
  char* result = NULL;
  
  switch( *p ) {
  case '<': /* email */ {
    int i;
    result = xmalloc( (strlen( pattern )+6)*sizeof(char) );
    if( !result ) return NULL;
    strcpy( result, "mail=" );
    strcat( result, p+1 );
    i = strlen(result)-1;
    if( result[i] == '>') result[i] = 0;
    /*log_debug( "%s\n", result );*/
    *base = 0;
    break;
  }
  case '/': /* Subject DN */
    *base = 1;
    result = xmalloc( strlen( pattern )*sizeof(char) );
    if( !result ) return NULL;
    strcpy( result, pattern+1 );
    break;
  case '#': /* Issuer DN */
    if( ++p && *p == '/' ) {
      /* Just issuer DN */
    } else {
      /* Serial number + issuer DN */
    }
    break;
  default: /* substring match */
    if( *p == '*' ) ++p;
    result = xmalloc( ( 2*strlen(p) + 19 ) * sizeof( char ) );
    /* |(cn=*xx*)(email=*xx*) */
    sprintf( result, "(|(cn=*%s*)(mail=*%s*))", p, p );
    /*sprintf( result, "cn=*%s*", p );*/ /* FIXME */
    *base = 0;
    break;
  };
  return result;
}

/* Iterate over certs matching pattern */
static LDAP*        _ld;
static STRLIST      _patterns;
static LDAPMessage* _result;
static const char* base = NULL;

int 
start_cert_fetch_ldap( STRLIST patterns, const struct ldap_server_s *server )
{
  /*
  const char* host = dirmngr_opt.ldapserver;
  int port = dirmngr_opt.ldapport;
  const char* user = dirmngr_opt.ldapuser;
  const char* pass = dirmngr_opt.ldappassword;
  */
  const char* host;
  int port;
  const char* user;
  const char* pass;
  base = "";
  //int rc;
  if( server ) {
    host = server->host;
    port = server->port;
    user = server->user;
    pass = server->pass;
    base = server->base;
    log_debug("Querying server %s\n", host );
  } else { abort(); }
  
  _patterns = patterns;
  _result = NULL;

#if 1
  {
    STRLIST sl;
    for( sl = _patterns; sl != NULL; sl = sl->next ){
      log_debug("pattern: %s\n", sl->d );
    }
  /*log_debug( "trying to fetch attr %s from %s/%s\n", attr, host, dn );*/
  }
#endif

  _ld = ldap_init( host, port );
  if( !_ld ) {
    perror( "Error in ldap_init" );
    return ASSUAN_No_CRL_For_Cert;
  }
  if( ldap_simple_bind_s( _ld, user, pass ) != LDAP_SUCCESS ) {
    ldap_perror( _ld, "Error in ldap_simple_bind_s" );
    return ASSUAN_No_CRL_For_Cert;
  }
  return ASSUAN_No_Error;
}

static char* find_next_pattern( int* use_as_base )
{
    const char* pattern;
    char* query = NULL;

    *use_as_base = 0;

    while( query == NULL ) {
      if( !_patterns ) break;
      pattern = _patterns->d;
      /*log_debug("Trying pattern %s\n", pattern );*/
      _patterns = _patterns->next;
      query = parse_one_pattern( pattern, use_as_base );
      /*log_debug("Got pattern %s\n", query );*/
    }
    return query;
}

static int
get_cert_ldap( char** value, size_t* valuelen )
{
  struct berval** values;
  char* dn = ldap_get_dn( _ld, _result );
  log_debug("Got DN=%s!!!\n", dn);  
  values = ldap_get_values_len( _ld, _result, CERTATTR );
  if( values ) {
    log_debug("Got userCertificate\n");
    *valuelen = values[0]->bv_len;
    *value = xmalloc( *valuelen * sizeof( char ) );
    memcpy( *value, values[0]->bv_val, *valuelen );
    ldap_value_free_len( values );
  } else {
    /* we might have a caCertificate */
    values = ldap_get_values_len( _ld, _result, CACERTATTR );
    if( values ) {
      log_debug("Got caCertificate\n");
      *valuelen = values[0]->bv_len;
      *value = xmalloc( *valuelen * sizeof( char ) );
      memcpy( *value, values[0]->bv_val, *valuelen );
      ldap_value_free_len( values );
    }
  }
  return ASSUAN_No_Error;
}

int 
fetch_next_cert_ldap( char** value, size_t* valuelen )
{
  int rc;
  int truncated = 0;
  char* attrs[] = {(char*)CERTATTR, (char*)CACERTATTR, NULL};
  log_debug("fetch_next_cert_ldap()\n");
  if( !_result ) {
    char* query;
    int use_as_base;
    if( !_patterns ) {
      log_debug("No more patterns\n");
      *value = 0;
      *valuelen = 0;
      return ASSUAN_No_Error;
    }
    query = find_next_pattern( &use_as_base );
    if( query ) {
      /*log_debug("got pattern\n");*/
      if( use_as_base ) {
	log_debug( "ldap_search_s with base = %s\n", query );
	rc = ldap_search_st( _ld, query, LDAP_SCOPE_SUBTREE, "objectClass=*", 
			    attrs, 0, &dirmngr_opt.ldaptimeout, &_result);
      } else {
	log_debug( "ldap_search_S with filter = %s\n", query );
	rc = ldap_search_st( _ld, base, LDAP_SCOPE_SUBTREE, query, 
			    attrs, 0, &dirmngr_opt.ldaptimeout, &_result);
      }
      if( rc != LDAP_SUCCESS ) {
	if( rc == LDAP_SIZELIMIT_EXCEEDED ) {
	  truncated = 1;
	  log_debug( "Warning, sizelimit of server exeeded\n");
	} else if( rc == LDAP_NO_SUCH_OBJECT ) {
	  /* ignore this */
	} else {
	  log_debug( "Error in line %s:%d\n", __FILE__, __LINE__ );
	  ldap_perror( _ld, "Error in ldap_search_s" );
	  return ASSUAN_General_Error;
	}
      }
      _result = ldap_first_entry( _ld, _result );
    } else {
      log_debug("No more good patterns\n");
      *value = 0;
      *valuelen = 0;
    }
  } else {
    /*_result = ldap_next_entry( _ld, _result );*/
  }
  if( _result ) {
    rc = get_cert_ldap( value, valuelen );
    if( rc ) return rc;
    _result = ldap_next_entry( _ld, _result );
    if( *value == NULL ) return fetch_next_cert_ldap( value, valuelen );
  } else {
    /*log_debug("crap!!!\n");*/
    if( _patterns ) return fetch_next_cert_ldap( value, valuelen );
  }

  if( !truncated ) return ASSUAN_No_Error;
  else return ASSUAN_DIRMNGR_Result_Truncated; /* FIXME */
}

int 
end_cert_fetch_ldap(void)
{
  _patterns = 0;
  _result = 0;
  _ld = 0;
  return ASSUAN_No_Error;
}
